unit Unit1;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs, ExtCtrls,
  StdCtrls, ComCtrls, Menus, Contnrs, Windows, WinSock2, MD5, Base64,
  OpenSSL, LazUTF8Classes;

type

  { TViewItem }

  TViewItem = class
  private
    FUidl: string;
    FHeader: TStrings;
    function _GetValue_Limited(Ident: string): string;
  public
    constructor Create(AUidl: string);
    destructor Destroy; override;
    function GetValue(Name: string): string;
    function GetSubject(): string;
    function GetFrom(): string;
  end;

  { TForm1 }

  TForm1 = class(TForm)
    Button_Update: TButton;
    ComboBox_Server: TComboBox;
    Label1: TLabel;
    LabeledEdit_Port: TLabeledEdit;
    LabeledEdit_UserID: TLabeledEdit;
    ListView1: TListView;
    MenuItem_ShowEmailHeader: TMenuItem;
    MenuItem_RetrEmail: TMenuItem;
    Panel1: TPanel;
    PopupMenu1: TPopupMenu;
    RadioGroup_Protocol: TRadioGroup;
    SaveDialog1: TSaveDialog;
    StatusBar1: TStatusBar;
    procedure Button_UpdateClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure ListView1Data(Sender: TObject; Item: TListItem);
    procedure ListView1DblClick(Sender: TObject);
    procedure MenuItem_RetrEmailClick(Sender: TObject);
    procedure MenuItem_ShowEmailHeaderClick(Sender: TObject);
    procedure PopupMenu1Popup(Sender: TObject);
  private
    { private declarations }
    FList: TObjectList;
    procedure DoProgress(SProgress: string);
    procedure WMUser_0(var Message: TMessage); message WM_USER + 0;
    procedure DoError(SError: string; AModalResult: Integer = mrAbort);
  protected
    function DoPop3Command(Sckt: TSocket; ACommand: string; var ARecv: string): Boolean; overload;
    function DoPop3Command(Sckt: TSocket; ACommand: string; ARecv: TStrings): Boolean; overload;
    function DoPop3_Top(): Boolean;
    function DoPop3_Retr(AViewItem: TViewItem): Boolean;
  protected
    function DoPop3Command(Ssl: OpenSSL.PSSL; ACommand: string; var ARecv: string): Boolean; overload;
    function DoPop3Command(Ssl: OpenSSL.PSSL; ACommand: string; ARecv: TStrings): Boolean; overload;
    function DoPop3Ssl_Top(): Boolean;
    function DoPop3Ssl_Retr(AViewItem: TViewItem): Boolean;
  public
    { public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.lfm}

uses
  CancelDlgUnit;

const
  CODEPAGE_UTF8 = 65001;
  CODEPAGE_SHIFTJIS = 932;
  CODEPAGE_ISO2022JP = 50220;
  CODEPAGE_EUCJP = 20932;
  CODEPAGE_DEFAULT = CODEPAGE_ISO2022JP;

procedure DEBUG_Message(S: string);
begin
  ShowMessage('DEBUG: ' + S);
end;

function Force_FileName(S: string; Max_Length: Integer = 16): string;
var
  WS: WideString;
  I: Integer;
begin
  WS := UTF8Decode(S);
  WS := Copy(WS, 1, Max_Length);
  for I := 1 to Length(WS) do
  case WS[I] of
  WideChar(0)..WideChar(31),
  WideChar('\'), WideChar('/'), WideChar(':'), WideChar('*'), WideChar('?'),
  WideChar('"'), WideChar('<'), WideChar('>'), WideChar('|'): WS[I] := '_';
  end;
  Result := UTF8Encode(WS);
end;

function Get_after_1stSpace(S: string): string; inline;
begin
  Delete(S, 1, Pos(' ', S));
  Result := Trim(S);
end;

function Get_before_1stSpace(S: string): string; inline;
begin
  S := Copy(S, 1, Pos(' ', S) - 1);
  Result := Trim(S);
end;

function multibyte_to_widestring(CodePage: UINT; S: AnsiString): WideString;
var
  dwFlags:DWORD; lpMultiByteStr:LPCSTR; cchMultiByte:longint; lpWideCharStr:LPWSTR; cchWideChar:longint;
var
  uI: UINT;
  pBuffer: PWideChar;
begin
  dwFlags := 0; // MB_PRECOMPOSED;
  lpMultiByteStr := PChar(S);
  cchMultiByte := -1; //  Length(S);
  lpWideCharStr := nil;
  cchWideChar := 0;
  uI := Windows.MultiByteToWideChar(
    CodePage, dwFlags, lpMultiByteStr, cchMultiByte, lpWideCharStr, cchWideChar);
  pBuffer := AllocMem((uI + 1) * sizeof(WideChar));
  try
    lpWideCharStr := pBuffer;
    cchWideChar := uI;

    uI := Windows.MultiByteToWideChar(
      CodePage, dwFlags, lpMultiByteStr, cchMultiByte, lpWideCharStr, cchWideChar);

    Result := WideString(pBuffer);
  finally
    FreeMem(pBuffer);
  end;
end;

function multibyte_to_utf8(CodePage: UINT; S: AnsiString): UTF8String;
begin
  if CodePage = 0 then
  begin
    if CODEPAGE_DEFAULT = CODEPAGE_UTF8 then
      Result := S
    else
      Result := UTF8Encode(multibyte_to_widestring(CODEPAGE_DEFAULT, S));
  end
  else if CodePage = CODEPAGE_UTF8 then
    Result := S
  else
    Result := UTF8Encode(multibyte_to_widestring(CodePage, S));
end;

function _Send(Sckt: TSocket; var sRecv: string; sSend: string; sEnd: string): Integer; overload;
const
  RECV_BUFFER_SIZE = $10000;
  TIME_OVER_MSEC = 5000;
var
  Recv_Start_TickCount: DWORD;
  MS: TMemoryStream;
  pRecvBuffer: Pointer;
  I, RecvReturnValue: Integer;
  S: string;
begin
  Result := -1;
  pRecvBuffer := AllocMem(RECV_BUFFER_SIZE);
  try
  MS := TMemoryStream.Create;
  try
    if(sSend <> '')then
      sSend := sSend + #13#10;

    if(sSend = '') or
      (WinSock2.send(Sckt, PChar(sSend), Length(sSend), 0) <> SOCKET_ERROR) then
    begin
      Sleep(0);
      Recv_Start_TickCount := GetTickCount();

      while TRUE do
      begin
        FillChar(PBYTE(pRecvBuffer)^, RECV_BUFFER_SIZE, 0);
        RecvReturnValue :=
          WinSock2.recv(Sckt, pRecvBuffer, RECV_BUFFER_SIZE, 0);
        Sleep(0);

        case RecvReturnValue of
        SOCKET_ERROR: // -1
          begin
            I := WinSock2.WSAGetLastError();
            if(I >= WSABASEERR)and(I <> WSAEWOULDBLOCK) then
            begin
              case I of
              WSAEINTR,
              WSAEFAULT,
              WSAEINPROGRESS,
              WSAENOTSOCK,
              WSAEMSGSIZE,
              WSAEOPNOTSUPP,
              WSAENETDOWN,
              WSAENETRESET,
              WSAECONNABORTED,
              WSAECONNRESET,
              WSAENOTCONN,
              WSAESHUTDOWN,
              WSAETIMEDOUT,
              WSANOTINITIALISED,
              WSAEINVAL:;
              end;
              break;
            end;// if(I >= WSABASEERR)and(I <> WSAEWOULDBLOCK) then
          end;
        0: break; // Socket Closed
        else
          begin
            MS.Write(PBYTE(pRecvBuffer)^, RecvReturnValue);
            // 正常に受信したのでタイムアウト判定の時刻を更新しておく
            Recv_Start_TickCount := GetTickCount();
          end;
        end; // case RecvReturnValue of

        // 終了判定はどうするのか？
        SetString(S, MS.Memory, MS.Size);
        if SameText(sEnd, Copy(S, Length(S) - Length(sEnd) + 1, Length(sEnd))) then
        begin
          sRecv := S;
          Result := 0;
          break;
        end;

        if (GetTickCount() - Recv_Start_TickCount) > TIME_OVER_MSEC then
        begin // Time Over !!!
          break;
        end;

      end; // while TRUE do

      // 不良終了の場合でも記録を返しておく
      if Result <> 0 then
        SetString(sRecv, MS.Memory, MS.Size);

    end;
  finally
    MS.Free;
  end;
  finally
    FreeMem(pRecvBuffer);
  end;
end;

function _Send(Ssl: PSSL; var sRecv: string; sSend: string; sEnd: string): Integer; overload;
const
  RECV_BUFFER_SIZE = $10000;
  TIME_OVER_MSEC = 5000;
var
  Recv_Start_TickCount: DWORD;
  MS: TMemoryStream;
  pRecvBuffer: Pointer;
  I, RecvReturnValue: Integer;
  S: string;
begin
  Result := -1;
  pRecvBuffer := AllocMem(RECV_BUFFER_SIZE);
  try
  MS := TMemoryStream.Create;
  try
    if(sSend <> '')then
      sSend := sSend + #13#10;

    if(sSend = '') or
      (OpenSSL.SslWrite(Ssl, PChar(sSend), Length(sSend)) > 0) then
    begin
      Sleep(0);
      Recv_Start_TickCount := GetTickCount();

      while TRUE do
      begin
        FillChar(PBYTE(pRecvBuffer)^, RECV_BUFFER_SIZE, 0);
        RecvReturnValue :=
          OpenSSL.SslRead(Ssl, pRecvBuffer, RECV_BUFFER_SIZE);
        Sleep(0);

        if RecvReturnValue < 0 then
        begin
          I := OpenSSL.SslGetError(Ssl, RecvReturnValue);
          break;
        end
        else if RecvReturnValue = 0 then
          // 受信データ量 ゼロ
        else // if RecvValue > 0 then
        begin
          MS.Write(PBYTE(pRecvBuffer)^, RecvReturnValue);
          // 正常に受信したのでタイムアウト判定の時刻を更新しておく
          Recv_Start_TickCount := GetTickCount();
        end;

        // 終了判定はどうするのか？
        SetString(S, MS.Memory, MS.Size);
        if SameText(sEnd, Copy(S, Length(S) - Length(sEnd) + 1, Length(sEnd))) or
          (sEnd = '') then
        begin
          sRecv := S;
          Result := 0;
          break;
        end;

        if (GetTickCount() - Recv_Start_TickCount) > TIME_OVER_MSEC then
        begin // Time Over !!!
          break;
        end;

      end; // while TRUE do

      // 不良終了の場合でも記録を返しておく
      if Result <> 0 then
        SetString(sRecv, MS.Memory, MS.Size);

    end;
  finally
    MS.Free;
  end;
  finally
    FreeMem(pRecvBuffer);
  end;
end;

function OpenFolder_with_FileSelection(Handle: HWND; SaveFileName: string): Boolean;
begin
  Result := ShellExecuteW(Handle,
    'open', 'explorer',
    PWideChar(UTF8Decode('/select, "' + SaveFileName +'"')),
    nil, SW_NORMAL) > 32;
end;

{ TViewItem }

constructor TViewItem.Create(AUidl: string);
begin
  FHeader := TStringList.Create;
  FUidl := AUidl;
end;

destructor TViewItem.Destroy;
begin
  FHeader.Free;
  inherited Destroy;
end;

function TViewItem.GetValue(Name: string): string;
var
  I, K: Integer;
  S: string;
begin
  Result := '?(' + Name + ')';
  for I := 0 to FHeader.Count - 1 do
  begin
    S := FHeader[I];
    if SameText(Name, Copy(S, 1, Length(Name))) then
    begin
      Result := Trim(Copy(S, Length(Name) + 1, MaxInt));
      for K := I + 1 to FHeader.Count - 1 do
      begin
        S := FHeader[K];
        if (Length(S) > 0)and(BYTE(S[1]) < BYTE(' ')) then
        begin
          S := Trim(S);
          if S <> '' then
            Result := Result + S;
        end
        else
          break;
      end;
      break;
    end;
  end;
end;

function TViewItem._GetValue_Limited(Ident: string): string;
var
  I, K: Integer;
  S: string;
  function _Decode_Limited(Src: string): string;
    function _Try(CodeName: string; CodePage: UINT; var SOut: string): Boolean;
    var
      S: string;
      I, K, L: Integer;
    begin
      S := '=?' + CodeName + '?B?';
      L := Length(S);
      I := Pos(S, UpperCase(Src));
      if I > 0 then
      begin
        K := Pos('?=', Copy(Src, I + L, MaxInt));
        if K > 0 then
          Inc(K, I + L - 1)
        else
          K := Length(Src) + 1;
        S := DecodeStringBase64(Copy(Src, I + L, K - I));
        SOut := Copy(Src, 1, I - 1) +
            Multibyte_to_UTF8(CodePage, S) +
            Copy(Src, K + 2, MaxInt);
        Result := TRUE;
      end
      else
        Result := FALSE;
    end;
  begin
    Src := Trim(Src);
    if _Try('UTF-8', CODEPAGE_UTF8, Result) then
    else if _Try('ISO-2022-JP', CODEPAGE_ISO2022JP, Result) then
    else if _Try('SHIFT_JIS', CODEPAGE_SHIFTJIS, Result) then
    else if _Try('EUC-JP', CODEPAGE_EUCJP, Result) then
    else
    begin
      // 実際には、他の文字コードがあったり、QuatedPrintableが使用されていたり、
      // もっと多様です。
      // "サンプルコード" として、手始めの人に手ごろな程度にとどめました。
      Result := Src;
    end;
  end;
begin
  // この部分のコードはサンプルと割りきって、かなり "手抜き" です。
  // そのため、正しくデコードできない場合がかなりあります。
  // "手抜き" を見抜いて補完して完成させる努力が必要になります。
  Result := '';
  for I := 0 to FHeader.Count - 1 do
  begin
    S := FHeader[I];
    if SameText(Ident, Copy(S, 1, Length(Ident))) then
    begin
      Result := _Decode_Limited(Copy(S, Length(Ident) + 1, MaxInt));
      for K := I + 1 to FHeader.Count - 1 do
      begin
        S := FHeader[K];
        if (Length(S) > 0)and(BYTE(S[1]) <= BYTE(' ')) then
        begin
          S := _Decode_Limited(S);
          if S <> '' then
            Result := Result + S;
        end
        else
          break;
      end;
      break;
    end;
  end;
end;

function TViewItem.GetSubject: string;
begin
  Result := _GetValue_Limited('Subject:');
end;

function TViewItem.GetFrom: string;
begin
  Result := _GetValue_Limited('From:');
end;

{ TForm1 }

procedure TForm1.FormCreate(Sender: TObject);
begin
  Caption := Application.Title;
  FList := TObjectList.create(TRUE);
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
  FList.Free;
end;

procedure TForm1.ListView1Data(Sender: TObject; Item: TListItem);
var
  ViewItem: TViewItem;
begin
  if(Item.Index >= 0)and(Item.Index < FList.Count)then
  begin
    ViewItem := TViewItem(FList[Item.Index]);
    if ViewItem.FHeader.Count > 0 then
    begin
      Item.Caption := ViewItem.GetSubject();
      Item.SubItems.Add(ViewItem.GetFrom());
      Item.SubItems.Add(ViewItem.GetValue('Date:'));
    end
    else
      Item.Caption := '* Message-ID:' + ViewItem.FUidl;
  end
  else
    Item.Caption := '(error)';
end;

procedure TForm1.ListView1DblClick(Sender: TObject);
var
  Shift: TShiftState;
begin
  Shift := GetKeyShiftState;
  if Shift = [] then
    MenuItem_RetrEmailClick(MenuItem_RetrEmail)
  else if Shift = [ssShift] then
    MenuItem_ShowEmailHeaderClick(MenuItem_RetrEmail);
end;

procedure TForm1.MenuItem_RetrEmailClick(Sender: TObject);
var
  I: Integer;
  ViewItem: TViewItem;
begin
  I := ListView1.ItemIndex;
  if(I >= 0)and(I < FList.Count)then
  begin
    ViewItem := TViewItem(FList[I]);
    if RadioGroup_Protocol.ItemIndex in [1, 2] then
      DoPop3Ssl_Retr(ViewItem)
    else
      DoPop3_Retr(ViewItem);
  end;
end;

procedure TForm1.MenuItem_ShowEmailHeaderClick(Sender: TObject);
var
  I: Integer;
begin
  I := ListView1.ItemIndex;
  if(I >= 0)and(I < FList.Count)then
    ShowMessage(TViewItem(FList[I]).FHeader.Text);
end;

procedure TForm1.PopupMenu1Popup(Sender: TObject);
var
  Check: Boolean;
  I: Integer;
begin
  Check := ListView1.ItemIndex >= 0;
  for I := 0 to PopupMenu1.Items.Count - 1 do
    PopupMenu1.Items[I].Enabled:= Check;
end;

function TForm1.DoPop3Command(Sckt: TSocket; ACommand: string; var ARecv: string): Boolean;
var
  S: string;
begin
  Result := FALSE;
  if _Send(Sckt, S, ACommand, #13#10) = 0 then
  begin
    if Copy(S, 1, 1) = '+' then
    begin
      ARecv := S;
      Result := TRUE;
    end;
  end;
end;

function TForm1.DoPop3Command(Sckt: TSocket; ACommand: string; ARecv: TStrings): Boolean;
var
  S: string;
begin
  Result := FALSE;
  if _Send(Sckt, S, ACommand, #13#10 + '.' + #13#10) = 0 then
  begin
    if SameText('+OK', Copy(S, 1, 3)) then
    begin
      ARecv.Text := S;
      Result := TRUE;
    end;
  end;
end;

function TForm1.DoPop3Command(Ssl: OpenSSL.PSSL; ACommand: string; var ARecv: string): Boolean;
var
  S: string;
begin
  Result := FALSE;
  if _Send(Ssl, S, ACommand, #13#10) = 0 then
  begin
    if Copy(S, 1, 1) = '+' then
    begin
      ARecv := S;
      Result := TRUE;
    end;
  end;
end;

function TForm1.DoPop3Command(Ssl: OpenSSL.PSSL; ACommand: string; ARecv: TStrings): Boolean;
var
  S: string;
begin
  Result := FALSE;
  if _Send(Ssl, S, ACommand, #13#10 + '.' + #13#10) = 0 then
  begin
    if Copy(S, 1, 1) = '+' then
    begin
      ARecv.Text := S;
      Result := TRUE;
    end;
  end;
end;

function TForm1.DoPop3_Top: Boolean;
var
  SPswd: string;
  wPort: WORD;
  pHE: PHostEnt;
  WSData: TWSAData;
  Sckt: TSocket;
  SA: TSockAddr;
  S, SRecv: string;
  I, K: Integer;
  UIDL: TStringList;
  R: TRect;
begin
  Result := FALSE;
  FList.Clear;
  ListView1.Items.Count := 0;

  DoProgress('Password');
  SPswd := PasswordBox('パスワード', '通信にSSLは用いません');
  if SPswd <> '' then
  begin
    // ポート番号の設定に TSpinEdit を使っても良かったのですが、
    // 空白なら標準値を設定するようにしてみました
    S := Trim(LabeledEdit_Port.Text);
    wPort := StrToIntDef(S, 110);

    DoProgress('ソケット通信を開始');
    if WSAStartup(2, WSData) = 0 then
    try
      DoProgress('サーバー名で接続先を探します');
      pHE := WinSock2.gethostbyname(PChar(ComboBox_Server.Text));
      if Assigned(pHE) then
      begin
        Sckt := WinSock2.socket(AF_INET, SOCK_STREAM, 0);
        if Sckt <> INVALID_SOCKET then
        begin
          FillChar(SA, sizeof(SA), 0);
          SA.sin_family := AF_INET;
          SA.sin_addr.s_addr := PDWORD(pHE^.h_addr_list[0])^;
          SA.sin_port := htons(wPort);

          DoProgress('サーバーに接続します');
          if WinSock2.connect(Sckt, SA, sizeof(SA)) = 0 then
          try
            // 接続メッセージを取得
            if DoPop3Command(Sckt, '', SRecv) then
            begin
              S := SRecv;
              I := Pos('<', S);
              if I > 0 then
                Delete(S, 1, I - 1);
              I := Pos('>', S);
              if I > 0 then
                S := Copy(S, 1, I);

              DoProgress('ログイン (APOP)');
              // SSLで暗号化されていないので、ログインはAPOP
              if DoPop3Command(Sckt, 'APOP ' + Self.LabeledEdit_UserID.Text + ' ' +
                   MD5.MD5Print(MD5.MD5String(S + SPswd)), SRecv) then
              begin
                UIDL := TStringList.Create;
                try
                  DoProgress('メールの識別コードリストを取得 (UIDL)');
                  if DoPop3Command(Sckt, 'UIDL', UIDL) then
                  begin
                    if UIDL.Count > 0 then
                    begin
                      //for I := 1 to UIDL.Count - 2 do
                      for I := UIDL.Count - 2 downto 1 do // 最新メールを先頭にする
                      begin
                        FList.Add(TViewItem.Create(UIDL[I]));
                      end;
                      ListView1.Items.Count := FList.Count;
                      ListView1.Repaint;

                      for I := 0 to FList.Count - 1 do
                      begin
                        S := TViewItem(FList[I]).FUidl;
                        K := Pos(' ', S);
                        S := Copy(S, 1, K - 1);
                        if IntToStr(StrToIntDef(S, -1)) = S then
                        begin
                          DoProgress('TOP ' + S + '(' + IntToStr(FList.Count) + ')');
                          if DoPop3Command(Sckt, 'TOP ' + S + ' 0', TViewItem(FList[I]).FHeader) then
                          begin
                            R := ListView1.Items[I].DisplayRect(drBounds);
                            if(R.Bottom > 0)and(R.Top < ListView1.ClientRect.Bottom)then
                            begin
                              ListView1.Invalidate;
                            end;
                          end
                          else
                          begin
                            DoError(S);
                            break;
                          end;
                        end
                        else DoError('extract Num from UIDL result');

                        Application.ProcessMessages;
                        if CancelDlg.ModalResult <> mrNone then
                          break;
                      end;//for I := 0 to FList.Count - 1 do

                      Result := CancelDlg.ModalResult = mrNone;
                    end
                    else DoError('UIDL.Count = ' + IntToStr(UIDL.Count));
                  end
                  else DoError('UIDL');
                finally
                  UIDL.Free;
                end;
              end
              else DoError('APOP');
            end
            else DoError('ERROR: 接続メッセージの取得');

            DoProgress('QUIT');
            if DoPop3Command(Sckt, 'QUIT', SRecv) = FALSE then
              DoError('QUIT');
          finally
            DoProgress('サーバーとの接続終了');
            WinSock2.closesocket(Sckt);
          end
          else DoError('connect, port = ' + IntToStr(wPort));
        end
        else DoError('socket');
      end
      else DoError('gethostbyname');
    finally
      DoProgress('ソケット通信を終了');
      WSACleanup();
    end
    else DoError('WSAStartUp');
  end
  else DoError('Password', mrCancel);

  DoProgress('');
  if (CancelDlg.ModalResult = mrNone)then
  begin
    if Result then
      CancelDlg.ModalResult := mrOK
    else
      CancelDlg.ModalResult := mrAbort;
  end;
end;

function TForm1.DoPop3_Retr(AViewItem: TViewItem): Boolean;
var
  SPswd: string;
  wPort: WORD;
  pHE: PHostEnt;
  WSData: TWSAData;
  Sckt: TSocket;
  SA: TSockAddr;
  S, SRecv: string;
  I: Integer;
  UIDL: TStringList;
  Retr: LazUTF8Classes.TStringListUTF8; // 日本語ファイル名による保存に対応するため
begin
  Result := FALSE;

  DoProgress('Password');
  SPswd := PasswordBox('パスワード', '通信にSSLは用いません');
  if SPswd <> '' then
  begin
    // ポート番号の設定に TSpinEdit を使っても良かったのですが、
    // 空白なら標準値を設定するようにしてみました
    S := Trim(LabeledEdit_Port.Text);
    wPort := StrToIntDef(S, 110);

    DoProgress('ソケット通信を開始');
    if WSAStartup(2, WSData) = 0 then
    try
      DoProgress('サーバー名で接続先を探します');
      pHE := WinSock2.gethostbyname(PChar(ComboBox_Server.Text));
      if Assigned(pHE) then
      begin
        Sckt := WinSock2.socket(AF_INET, SOCK_STREAM, 0);
        if Sckt <> INVALID_SOCKET then
        begin
          FillChar(SA, sizeof(SA), 0);
          SA.sin_family := AF_INET;
          SA.sin_addr.s_addr := PDWORD(pHE^.h_addr_list[0])^;
          SA.sin_port := htons(wPort);

          DoProgress('サーバーに接続します');
          if WinSock2.connect(Sckt, SA, sizeof(SA)) = 0 then
          try
            // 接続メッセージを取得
            if DoPop3Command(Sckt, '', SRecv) then
            begin
              S := SRecv;
              I := Pos('<', S);
              if I > 0 then
                Delete(S, 1, I - 1);
              I := Pos('>', S);
              if I > 0 then
                S := Copy(S, 1, I);

              DoProgress('ログイン (APOP)');
              // SSLで暗号化されていないので、ログインはAPOP
              if DoPop3Command(Sckt, 'APOP ' + Self.LabeledEdit_UserID.Text + ' ' +
                   MD5.MD5Print(MD5.MD5String(S + SPswd)), SRecv) then
              begin
                UIDL := TStringList.Create;
                try
                  DoProgress('メールの識別コードリストを取得 (UIDL)');
                  if DoPop3Command(Sckt, 'UIDL', UIDL) then
                  begin
                    if UIDL.Count > 0 then
                    begin
                      // UIDLのIDが一致する項目を探す
                      for I := 1 to UIDL.Count - 2 do
                      begin
                        if Get_after_1stSpace(AViewItem.FUidl) =
                           Get_after_1stSpace(UIDL[I]) then
                        begin
                          // メールデータを取得してファイル保存する
                          Retr := LazUTF8Classes.TStringListUTF8.Create;
                          try
                            S := Get_before_1stSpace(UIDL[I]);
                            DoProgress('RETR ' + S);
                            if DoPop3Command(Sckt, 'Retr ' + S, Retr) then
                            begin
                              if Retr.Count > 2 then
                              begin
                                Retr.Delete(0);
                                Retr.Delete(Retr.Count - 1);

                                S := Force_FileName(AViewItem.GetSubject);
                                SaveDialog1.FileName := S + SaveDialog1.DefaultExt;
                                if SaveDialog1.Execute then
                                begin
                                  Retr.SaveToFile(SaveDialog1.FileName);
                                  if Application.MessageBox('保存したフォルダを開きますか？',
                                    '電子メールファイル(*.eml)の保存',
                                    MB_ICONQUESTION or MB_YESNO) = IDYES then
                                    OpenFolder_with_FileSelection(Handle, SaveDialog1.FileName);
                                end;

                                Result := CancelDlg.ModalResult = mrNone;
                              end
                              else DoError('Retr.Count = ' + IntToStr(Retr.Count));
                            end
                            else DoError('Retr');

                          finally
                            Retr.Free;
                          end;
                          break;
                        end;
                      end;//for I := 1 to UIDL.Count - 2 do
                    end
                    else DoError('UIDL.Count = ' + IntToStr(UIDL.Count));
                  end
                  else DoError('UIDL');
                finally
                  UIDL.Free;
                end;
              end
              else DoError('APOP');
            end
            else DoError('ERROR: 接続メッセージの取得');

            DoProgress('QUIT');
            if DoPop3Command(Sckt, 'QUIT', SRecv) = FALSE then
              DoError('QUIT');
          finally
            DoProgress('サーバーとの接続終了');
            WinSock2.closesocket(Sckt);
          end
          else DoError('connect, port = ' + IntToStr(wPort));
        end
        else DoError('socket');
      end
      else DoError('gethostbyname');
    finally
      DoProgress('ソケット通信を終了');
      WSACleanup();
    end
    else DoError('WSAStartUp');
  end
  else DoError('Password', mrCancel);

  DoProgress('');
  if (CancelDlg.ModalResult = mrNone)then
  begin
    if Result then
      CancelDlg.ModalResult := mrOK
    else
      CancelDlg.ModalResult := mrAbort;
  end;
end;

function TForm1.DoPop3Ssl_Top: Boolean;
var
  SPswd: string;
  wPort: WORD;
  pHE: PHostEnt;
  WSData: TWSAData;
  Sckt: TSocket;
  SA: TSockAddr;
  pCtx: OpenSSL.PSSL_CTX;
  Ssl: OpenSSL.PSSL;
  I, K: Integer;
  S, SRecv: string;
  UIDL: TStringList;
  R: TRect;
begin
  Result := FALSE;
  FList.Clear;
  ListView1.Items.Count := 0;

  DoProgress('SSL の初期化');
  OpenSSL.InitSSLEAInterface(FALSE);
  OpenSSL.InitSSLInterface();
  OpenSSL.InitLibeaInterface();
  if OpenSSL.IsSSLloaded then
  begin
    DoProgress('Password');
    SPswd := PasswordBox('パスワード', 'OpenSSLで暗号化通信を試みます');
    if SPswd <> '' then
    begin
      // ポート番号の設定に TSpinEdit を使っても良かったのですが、
      // 空白なら標準値を設定するようにしてみました
      S := Trim(LabeledEdit_Port.Text);
      wPort := StrToIntDef(S, 995);

      DoProgress('ソケット通信を開始');
      if WSAStartup(2, WSData) = 0 then
      try
        DoProgress('サーバー名で接続先を探します');
        pHE := WinSock2.gethostbyname(PChar(ComboBox_Server.Text));
        if Assigned(pHE) then
        begin
          Sckt := WinSock2.socket(AF_INET, SOCK_STREAM, 0);
          if Sckt <> INVALID_SOCKET then
          begin
            FillChar(SA, sizeof(SA), 0);
            SA.sin_family := AF_INET;
            SA.sin_addr.s_addr := PDWORD(pHE^.h_addr_list[0])^;
            SA.sin_port := htons(wPort);

            DoProgress('サーバーに接続します');
            if WinSock2.connect(Sckt, SA, sizeof(SA)) = 0 then
            try
              pCtx := OpenSSL.SslCtxNew(SslMethodV23());
              try
                if OpenSSL.SslCTXCtrl(pCtx, SSL_CTRL_OPTIONS, SSL_OP_NO_SSLv2, nil)
                  <> 0 then
                else DoError('SSL_Stx_Ctrl', mrNone);

                Ssl := OpenSSL.SslNew(pCtx);
                try
                  if OpenSSL.SslSetFd(Ssl, Sckt) <> 0 then
                  begin
                    DoProgress('Connecting (SSL)');
                    if OpenSSL.SslConnect(Ssl) > 0 then
                    begin
                      DoProgress('SSL暗号化の種類: ' + SSLCipherGetName(SSLGetCurrentCipher(Ssl)));
                      repeat
                        // 接続メッセージを取得
                        if DoPop3Command(Ssl, '', SRecv) then
                        begin
                          S := SRecv;

                          if RadioGroup_Protocol.ItemIndex = 1 then
                          begin
                            DoProgress('ログイン (Plain)');
                            if DoPop3Command(Ssl, 'USER ' + LabeledEdit_UserID.Text, SRecv) = FALSE then
                            begin
                              DoError('USER');
                              break;
                            end;

                            if DoPop3Command(Ssl, 'PASS ' + sPswd, SRecv) = FALSE then
                            begin
                              DoError('PASS');
                              break;
                            end;
                          end
                          else // if APOP
                          begin
                            DoProgress('ログイン (APOP)');
                            I := Pos('<', S);
                            if I > 0 then
                              Delete(S, 1, I - 1);
                            I := Pos('>', S);
                            if I > 0 then
                              S := Copy(S, 1, I);

                            if DoPop3Command(Ssl, 'APOP ' + LabeledEdit_UserID.Text + ' ' +
                                 MD5.MD5Print(MD5.MD5String(S + SPswd)), SRecv) = FALSE then
                            begin
                              DoError('APOP');
                              break;
                            end;
                          end;

                          UIDL := TStringList.Create;
                          try
                            DoProgress('メールの識別コードリストを取得 (UIDL)');
                            if DoPop3Command(Ssl, 'UIDL', UIDL) then
                            begin
                              if UIDL.Count > 0 then
                              begin
                                //for I := 1 to UIDL.Count - 2 do
                                for I := UIDL.Count - 2 downto 1 do // 最新メールを先頭にする
                                begin
                                  FList.Add(TViewItem.Create(UIDL[I]));
                                end;
                                ListView1.Items.Count := FList.Count;
                                ListView1.Repaint;

                                for I := 0 to FList.Count - 1 do
                                begin
                                  S := TViewItem(FList[I]).FUidl;
                                  K := Pos(' ', S);
                                  S := Copy(S, 1, K - 1);
                                  if IntToStr(StrToIntDef(S, -1)) = S then
                                  begin
                                    DoProgress('TOP ' + S + '(' + IntToStr(FList.Count) + ')');
                                    if DoPop3Command(Ssl, 'TOP ' + S + ' 0', TViewItem(FList[I]).FHeader) then
                                    begin
                                      R := ListView1.Items[I].DisplayRect(drBounds);
                                      if(R.Bottom > 0)and(R.Top < ListView1.ClientRect.Bottom)then
                                      begin
                                        ListView1.Invalidate;
                                      end;
                                    end
                                    else
                                    begin
                                      DoError(S);
                                      break;
                                    end;
                                  end
                                  else DoError('extract Num from UIDL result');

                                  Application.ProcessMessages;
                                  if CancelDlg.ModalResult <> mrNone then
                                    break;
                                end;//for I := 0 to FList.Count - 1 do

                                Result := CancelDlg.ModalResult = mrNone;
                              end
                              else DoError('UIDL.Count = ' + IntToStr(UIDL.Count));
                            end
                            else DoError('UIDL');
                          finally
                            UIDL.Free;
                          end;

                        end
                        else DoError('ERROR: 接続メッセージの取得');
                      until TRUE;

                      // QUITコマンド送信
                      DoProgress('QUIT');
                      if DoPop3Command(Ssl, 'QUIT', SRecv) = FALSE then
                        DoError('QUIT');
                    end
                    else DoError('OpenSSL.SslConnect');
                  end
                  else DoError('OpenSSL.SslSetFd');
                finally
                  OpenSSL.SslFree(Ssl);
                end;
              finally
                OpenSSL.SslCtxFree(pCtx);
              end;
            finally
              DoProgress('サーバーとの接続終了');
              WinSock2.closesocket(Sckt);
            end
            else DoError('connect, port = ' + IntToStr(wPort));
          end
          else DoError('socket');
        end
        else DoError('gethostbyname');
      finally
        DoProgress('ソケット通信を終了');
        WSACleanup();
      end
      else DoError('WSAStartUp');
    end
    else DoError('Password', mrCancel);
  end
  else DoError('OpenSSL が利用できません');

  DoProgress('');
  if (CancelDlg.ModalResult = mrNone)then
  begin
    if Result then
      CancelDlg.ModalResult := mrOK
    else
      CancelDlg.ModalResult := mrAbort;
  end;
end;

function TForm1.DoPop3Ssl_Retr(AViewItem: TViewItem): Boolean;
var
  SPswd: string;
  wPort: WORD;
  pHE: PHostEnt;
  WSData: TWSAData;
  Sckt: TSocket;
  SA: TSockAddr;
  pCtx: OpenSSL.PSSL_CTX;
  Ssl: OpenSSL.PSSL;
  I: Integer;
  S, SRecv: string;
  UIDL: TStringList;
  Retr: LazUTF8Classes.TStringListUTF8; // 日本語ファイル名による保存に対応するため
begin
  Result := FALSE;

  DoProgress('SSL の初期化');
  OpenSSL.InitSSLEAInterface(FALSE);
  OpenSSL.InitSSLInterface();
  OpenSSL.InitLibeaInterface();
  if OpenSSL.IsSSLloaded then
  begin
    DoProgress('Password');
    SPswd := PasswordBox('パスワード', 'OpenSSLで暗号化通信を試みます');
    if SPswd <> '' then
    begin
      // ポート番号の設定に TSpinEdit を使っても良かったのですが、
      // 空白なら標準値を設定するようにしてみました
      S := Trim(LabeledEdit_Port.Text);
      wPort := StrToIntDef(S, 995);

      DoProgress('ソケット通信を開始');
      if WSAStartup(2, WSData) = 0 then
      try
        DoProgress('サーバー名で接続先を探します');
        pHE := WinSock2.gethostbyname(PChar(ComboBox_Server.Text));
        if Assigned(pHE) then
        begin
          Sckt := WinSock2.socket(AF_INET, SOCK_STREAM, 0);
          if Sckt <> INVALID_SOCKET then
          begin
            FillChar(SA, sizeof(SA), 0);
            SA.sin_family := AF_INET;
            SA.sin_addr.s_addr := PDWORD(pHE^.h_addr_list[0])^;
            SA.sin_port := htons(wPort);

            DoProgress('サーバーに接続します');
            if WinSock2.connect(Sckt, SA, sizeof(SA)) = 0 then
            try
              pCtx := OpenSSL.SslCtxNew(SslMethodV23());
              try
                if OpenSSL.SslCTXCtrl(pCtx, SSL_CTRL_OPTIONS, SSL_OP_NO_SSLv2, nil)
                  <> 0 then
                else DoError('SSL_Stx_Ctrl', mrNone);

                Ssl := OpenSSL.SslNew(pCtx);
                try
                  if OpenSSL.SslSetFd(Ssl, Sckt) <> 0 then
                  begin
                    DoProgress('Connecting (SSL)');
                    if OpenSSL.SslConnect(Ssl) > 0 then
                    begin
                      DoProgress('SSL暗号化の種類: ' + SSLCipherGetName(SSLGetCurrentCipher(Ssl)));
                      repeat
                        // 接続メッセージを取得
                        if DoPop3Command(Ssl, '', SRecv) then
                        begin
                          S := SRecv;

                          if RadioGroup_Protocol.ItemIndex = 1 then
                          begin
                            DoProgress('ログイン (Plain)');
                            if DoPop3Command(Ssl, 'USER ' + LabeledEdit_UserID.Text, SRecv) = FALSE then
                            begin
                              DoError('USER');
                              break;
                            end;

                            if DoPop3Command(Ssl, 'PASS ' + sPswd, SRecv) = FALSE then
                            begin
                              DoError('PASS');
                              break;
                            end;
                          end
                          else // if APOP
                          begin
                            DoProgress('ログイン (APOP)');
                            I := Pos('<', S);
                            if I > 0 then
                              Delete(S, 1, I - 1);
                            I := Pos('>', S);
                            if I > 0 then
                              S := Copy(S, 1, I);

                            if DoPop3Command(Ssl, 'APOP ' + LabeledEdit_UserID.Text + ' ' +
                                 MD5.MD5Print(MD5.MD5String(S + SPswd)), SRecv) = FALSE then
                            begin
                              DoError('APOP');
                              break;
                            end;
                          end;

                          UIDL := TStringList.Create;
                          try
                            DoProgress('メールの識別コードリストを取得 (UIDL)');
                            if DoPop3Command(Ssl, 'UIDL', UIDL) then
                            begin
                              if UIDL.Count > 0 then
                              begin
                                // UIDLのIDが一致する項目を探す
                                for I := 1 to UIDL.Count - 2 do
                                begin
                                  if Get_after_1stSpace(AViewItem.FUidl) =
                                     Get_after_1stSpace(UIDL[I]) then
                                  begin
                                    // メールデータを取得してファイル保存する
                                    Retr := LazUTF8Classes.TStringListUTF8.Create;
                                    try
                                      S := Get_before_1stSpace(UIDL[I]);
                                      DoProgress('Retr ' + S);
                                      if DoPop3Command(Ssl, 'Retr ' + S, Retr) then
                                      begin
                                        if Retr.Count > 2 then
                                        begin
                                          Retr.Delete(0);
                                          Retr.Delete(Retr.Count - 1);

                                          S := Force_FileName(AViewItem.GetSubject);
                                          SaveDialog1.FileName := S + SaveDialog1.DefaultExt;
                                          if SaveDialog1.Execute then
                                          begin
                                            Retr.SaveToFile(SaveDialog1.FileName);
                                            if Application.MessageBox('保存したフォルダを開きますか？',
                                              '電子メールファイル(*.eml)の保存',
                                              MB_ICONQUESTION or MB_YESNO) = IDYES then
                                              OpenFolder_with_FileSelection(Handle, SaveDialog1.FileName);
                                          end;

                                          Result := CancelDlg.ModalResult = mrNone;
                                        end
                                        else DoError('Retr.Count = ' + IntToStr(Retr.Count));
                                      end
                                      else DoError('Retr');
                                    finally
                                      Retr.Free;
                                    end;
                                    break;
                                  end;
                                end;//for I := 1 to UIDL.Count - 2 do
                              end
                              else DoError('UIDL.Count = ' + IntToStr(UIDL.Count));
                            end
                            else DoError('UIDL');
                          finally
                            UIDL.Free;
                          end;

                        end
                        else DoError('ERROR: 接続メッセージの取得');
                      until TRUE;

                      // QUITコマンド送信
                      DoProgress('QUIT');
                      if DoPop3Command(Ssl, 'QUIT', SRecv) = FALSE then
                        DoError('QUIT');
                    end
                    else DoError('OpenSSL.SslConnect');
                  end
                  else DoError('OpenSSL.SslSetFd');
                finally
                  OpenSSL.SslFree(Ssl);
                end;
              finally
                OpenSSL.SslCtxFree(pCtx);
              end;
            finally
              DoProgress('サーバーとの接続終了');
              WinSock2.closesocket(Sckt);
            end
            else DoError('connect, port = ' + IntToStr(wPort));
          end
          else DoError('socket');
        end
        else DoError('gethostbyname');
      finally
        DoProgress('ソケット通信を終了');
        WSACleanup();
      end
      else DoError('WSAStartUp');
    end
    else DoError('Password', mrCancel);
  end
  else DoError('OpenSSL が利用できません');

  DoProgress('');
  if (CancelDlg.ModalResult = mrNone)then
  begin
    if Result then
      CancelDlg.ModalResult := mrOK
    else
      CancelDlg.ModalResult := mrAbort;
  end;
end;

procedure TForm1.DoProgress(SProgress: string);
begin
  CancelDlg.CommentLabel.Caption := SProgress;
  StatusBar1.SimpleText := SProgress;
  CancelDlg.CommentLabel.Invalidate;
  StatusBar1.Invalidate;
  Application.ProcessMessages;
end;

procedure TForm1.WMUser_0(var Message: TMessage);
begin
  if RadioGroup_Protocol.ItemIndex in [1, 2] then
    DoPop3Ssl_Top()
  else
    DoPop3_Top();
end;

procedure TForm1.DoError(SError: string; AModalResult: Integer);
var
  SType: string;
begin
  if CancelDlg.Visible then
    CancelDlg.ModalResult:= AModalResult;
  case AModalResult of
  mrCancel: SType := 'Cancel: ';
  mrAbort: SType := 'ERROR: ';
  else SType := 'ERROR: ';
  end;
  ShowMessage(SType + SError);
end;

procedure TForm1.Button_UpdateClick(Sender: TObject);
begin
  PostMessage(Handle, WM_USER + 0, 0, 0);
  case CancelDlgUnit.CancelDlg.ShowModal of
  mrCancel: ShowMessage('キャンセルされました');
  mrAbort: ShowMessage('Abortされました');
  end;
end;

end.

